from pig_ui.constants import *
from pig_ui.ui.ux_element import UXRenderer, UXWrapper, UIELEMENT_DEFAULT, UIELEMENT_TEXT, UXText, UXRect
from pig_ui.ui.ui_manager import UIM
from pig_ui.events import Events

class UIGroup:
    gid = 0
    def __init__(self, group_name: str):
        self.group_name = group_name
        UIGroup.gid += 1

UI_DEFAULT_GROUP = UIGroup('default')

ANCHORS = {'t': 0.0, 'c': 0.5, 'b': 1.0, 'l': 0.0, 'r': 1.0}

class UIElement: ...
class UIElement:
    """
    The Base-Class for all UIElements
    ***
    A UIElement supports by default the following events & its corresponding callbacks:
    |key|callback name|
    |---|---|
    |cb_lclick|left click|
    |cb_dclick|double click|
    |cb_rclick|right click|
    |cb_unclick|release|
    |cb_hover|hover|
    |cb_unhover|unhover|
    |cb_drag|drag|
    |cb_wheel|scroll|
    |cb_keypress|keypress|
    
    ## Params
    #### app
    :key key1:
    The UI Element needs this to:
    - blit to the window
    - retreive events
    
    #### pos
    This is the Offset to the parent. If no parent is given, use pos+(0,0) as start.

    #### size
    Used to determine the Default UXElement size & the bounding box

    > [!NOTE]
    > You should not change the size in the lifetime, only in the init state.
    > You must change the UXElement sizes manually!
    
    #### ux
    Must be a UXWrapper Object.
    
    `ux` is used for styling the UIElement, for more info: look into the UXElement Docs.
    
    #### kwargs
    event
    group
    layer
    o_layer
    visible
    anchor
    parent
    draggable
    blocked
    click_offset
    uid
     
    ## Attributes
    
    
    """
    #! Fix click in non hover state -> sliding into uie will count as click
    """
    The Base Class for UIElements
    """
    __uid = 0
    def __init__(self,
                 app,
                 pos: Vector2,
                 size: Vector2,
                 ux: UXWrapper | None = None,
                 draggable: bool = False,
                 **kwargs):
        
        self.app = app
        if ux is None:
            self.ux = UXWrapper(UIELEMENT_DEFAULT)
        else: 
            self.ux = ux
        self.event: Events = self.app.events
        self.pos = pos
        self.size = size
        self.draggable = draggable
        self.group = kwargs.get('group',UI_DEFAULT_GROUP)
        
        self.layer = kwargs.get('layer',0)
        self.o_layer = kwargs.get('o_layer',0)
        self.visible = kwargs.get('visible', True)
        self.anchor = kwargs.get('anchor', 'cc')
        self.parent = kwargs.get('parent',None)
        
        self.callback_hover = kwargs.get('cb_hover',lambda x: print(f"hover: {x}"))
        self.callback_unhover = kwargs.get('cb_unhover',lambda x: print(f"unhover: {x}"))
        self.callback_lclick = kwargs.get('cb_lclick',lambda x: print(f"LEFT: {x}"))
        self.callback_unclick = kwargs.get('cb_unclick',lambda x: None)
        self.callback_rclick = kwargs.get('cb_rclick',lambda x: print(f"RIGHT: {x}"))
        self.callback_dclick = kwargs.get('cb_dclick',lambda x: print(f"DOUBLE: {x}"))
        self.callback_drag = kwargs.get('cb_drag',lambda x: print(f"DRAG: {x}"))
        self.callback_wheel = kwargs.get('cb_wheel',lambda x: print(f"WHEEL: {x}"))
        self.callback_keypress = kwargs.get('cb_keypress',lambda x: print(f"KEY: {x}"))
        
        self.blocked = False
        self.click_offset = Vector2(0,0)
        
        UIElement.__uid += 1
        self.uid = UIElement.__uid
        UIM.add_to_queue(self)
    
    @property
    def rect(self) -> Rect:
        """
        Returns the Rect from this UI Object.
        """
        return Rect(self.pos.x, self.pos.y, self.size.x, self.size.y)
    
    @property
    def hover(self) -> bool:
        x, y = self.abs_offset
        w,h = self.size.x,self.size.y
        g,l = self.event.MOUSE_POS
        return g > x and l > y and g < x + w and l < y + h
    
    @property
    def image(self) -> Surface:
        """
        Returns the current Image for this UIElement.
        Will be a blank 1 by 1 Surface if nothing set!
        """
        if not hasattr(self,'surface'):
            self.surface = Surface((1,1))
        return self.surface
    
    @property
    def anchor_offset(self) -> Vector2:
        """
        Calculates the offset for the anchor: Center, LEFT etc.
        """
        #TODO add offset code here
        x,y = self.anchor
        return Vector2(ANCHORS[x] * self.size.x, ANCHORS[y] * self.size.y)
    
    @property
    def parent_offset(self) -> Vector2:
        """
        Calculates the parent offset to the `regular` position of the `UIElement`
        """
        offset = Vector2(0,0)
        parent = self.parent
        
        if parent is not None:
            offset += parent.abs_offset
            
        return offset
    
    @property
    def abs_offset(self) -> Vector2:
        """
        The absolute offset: anchor + parents + pos
        """
        return self.anchor_offset + self.parent_offset + self.pos
    
    @property
    def root_parent(self) -> None | UIElement:
        """
        Gets the main / root parent of this object.
        Can only be None if the first upper parent is None.
        
        """
        
        parent = self.parent
        backup = self.parent
        i = 0
        while parent.parent is not None:
            parent = parent.parent
            
        return parent if i > 0 else backup
    
    @property
    def get_internal_mouse_pos(self) -> Vector2:
        return self.event.MOUSE_POS - self.abs_offset
    
    def mouse_interaction(self):
        
        if self.event.MOUSE_LEFT and not self.hover and not self.blocked:
            self.callback_unclick(self)
        
        # Click only one time, self.last_frame_hold to time
        if self.event.MOUSE_LEFT and self.is_free:
            self.callback_lclick(self)
            self.ux.set_mode(2)
        
        # UI Double Clicked
        if self.event.DOUBLE_CLICK and self.is_free: 
            self.callback_dclick(self)
            self.ux.set_mode(2)
            
        # Right click
        if self.event.MOUSE_RIGHT and self.is_free:
            self.callback_rclick(self)
            
        # A key was pressed while hover
        if self.event.KEYDOWN and self.event.KEYS and self.is_free:
            self.callback_keypress(self)
        
        if self.event.WHEEL and self.is_free:
            self.callback_wheel(self)
        
    def mouse_interaction_ex(self):
        # Dragging
        if self.event.MOUSE_MIDDLE and self.is_free:
            self.is_dragging = True
            self.click_offset = self.pos - self.event.MOUSE_POS
            self.callback_drag(self)
            self.ux.set_mode(2)
        
    def keyboard_interaction(self): # Used for Shortcuts
        ...
    def keyboard_interaction_ex(self): # Used for TextInputSystems
        ...
    
    @property
    def is_free(self) -> bool:
        return self.hover and not self.blocked
    
    def update(self): 
        # self.last_frame_hold to False if not pressed anymore
        # self.last_frame_hold must have a multiframe puffer preventing unwanted drags.
        # So self.last_frame_hold must be >= last_time + 0.2 and after that the self.dragging will be set to true
        
        
        #! If a object is dragged, only draw other update, DO NOT UPDATE!
        
        # Updating position on dragging if enabled
        if self.event.MOUSE_MIDDLE and self.is_dragging:
            self.pos = self.event.MOUSE_POS + self.click_offset
        
        # Resets the ability to use the UIE
        # Currently it is pretty simple and will prevent from pressing Mouse 1 after Mouse 3 etc.
        if not self.event.MOUSE_LEFT \
            and not self.event.MOUSE_RIGHT \
            and not self.event.MOUSE_MIDDLE \
            and not self.event.KEYS \
            and not self.event.DOUBLE_CLICK\
            and not self.event.WHEEL:
            self.is_dragging = False
            self.blocked = False
            self.click_offset = Vector2(0,0)
            UIM.blocked = -1
            return False
        
        self.mouse_interaction()

        if self.draggable:
            self.mouse_interaction_ex()
        
        self.keyboard_interaction()
        
        self.blocked = any([self.event.MOUSE_LEFT,self.event.MOUSE_RIGHT,self.event.MOUSE_MIDDLE,self.event.KEYS,self.event.DOUBLE_CLICK,self.event.WHEEL]) and self.hover
        return self.blocked
    
    def draw(self):
        self.app.window.blit(self.image,self.abs_offset)
        PG.draw.rect(self.app.window, (255,0,0), (self.abs_offset.x, self.abs_offset.y, self.size.x, self.size.y))

        if not self.blocked and UIM.blocked == -1:
            self.ux.set_mode(0)
        self.ux.draw(self.app.window,self.abs_offset)
        
    def destroy(self):
        UIM.remove_from_queue(self)
        del self



class Anchors:
    TOP_LEFT = 'tl'
    TOP_CENTER = 'tc'
    TOP_RIGHT = 'tr'
    
    CCENTER = 'cc'
    CENTER_LEFT = 'cl'
    CENTER_RIGHT = 'cr'
    
    BOTTOM_LEFT = 'bl'
    BOTTOM_RIGHT = 'br'
    BOTTOM_CENTER = 'bc'


from pig_ui.ui.ui_element import UIElement
from pig_ui.ui.ui_textinput import UITextInput
from pig_ui.ui.ux_element import UXImage, UXWrapper
from numpy import zeros, uint8
from numpy.typing import NDArray
from colorsys import rgb_to_hsv
from pig_ui.constants import *
from math import floor
from pig_ui.ui.ux_element import UXRect, UXText

SET_BUTTON = [
    [UXRect(-1,Color('#484848'),size=Vector2(128,16))],
    [UXRect(-1,Color('#969696'),size=Vector2(128,16))],
    [UXRect(-1,Color('#ffffff'),size=Vector2(128,16))],
    [UXRect(-1,Color('#000000'),size=Vector2(128,16))]
]


def convert_hsv_to_rgb(H: float, S: float, V: float) -> tuple[int, int, int]:
    """
    Converts HSV to RGB
    """
    if S == 0.0:
        val = int(V * 255)
        return (val, val, val)
    i = floor(H * 6)
    f = H * 6 - i
    x = V * (1 - S)
    y = V * (1 - f * S)
    z = V * (1 - (1 - f) * S)
    match i % 6:
        case 0: r, g, b = V, z, x
        case 1: r, g, b = y, V, x
        case 2: r, g, b = x, V, z
        case 3: r, g, b = x, y, V
        case 4: r, g, b = z, x, V
        case 5: r, g, b = V, x, y
        case _: r, g, b = 0, 0, 0 # H >= 1.0
    
    return (int(r * 255), int(g * 255), int(b * 255))

def color_rect_array(hue: float, size: int = 256) -> NDArray:
    """
    Gives a pixel_array of a color_palette based of the hue, used for the colorpicker
    
    This is not optimized:
        - only drawing pixels needed, not all & convert
    """
    pixel_array = zeros(shape=(256,256,3),dtype=uint8)
    for x in range(256):
        for y in range(256):
            pixel_array[y, x] = convert_hsv_to_rgb(hue,x / 255,(255 - y) / 255)
    return pixel_array

def color_rect(hue: float, size: int = 256) -> Surface:
    """
    Gives a Surface of a color_palette based of the hue, used for the colorpicker
    """
    return PG.transform.scale(make_surface(color_rect_array(hue, size)),(size, size))

def color_line(size: Vector2 = Vector2(256, 16)) -> Surface:
    """
    Gives a pixel_array of a color_line for selecting the hue of the color_rect, used for the colorpicker
    
    This is not optimized:
        - only drawing pixels needed, not all & convert
    """
    pixel_array = zeros(shape=(256,16,3))
    for x in range(256):
        for y in range(16):
            pixel_array[x, y] = convert_hsv_to_rgb(x/255,1.0,1.0)
    return PG.transform.scale(make_surface(pixel_array),size)

class UIColorPickerRect(UIElement):
    def __init__(self, app, pos, size, ux = None, draggable = False, **kwargs):
        
        super().__init__(app, pos, size, ux, draggable, **kwargs)
    
    def draw(self):
        super().draw()
        PG.draw.circle(
                self.app.window,
                (self.parent.color.b, self.parent.color.g, self.parent.color.r),
                self.abs_offset + self.parent.color_pos,
                15,
                1
            )
class UIColorPickerLine(UIElement):
    def __init__(self, app, pos, size, ux = None, draggable = False, **kwargs):
        
        super().__init__(app, pos, size, ux, draggable, **kwargs)
    
    def draw(self):
        super().draw()
        col = Color(self.parent.ux_color_line.image.get_at(Vector2(self.parent.hue_pos,0)))
        PG.draw.line(
                self.app.window,
                (col.b, col.g, col.r),
                self.abs_offset + Vector2(self.parent.hue_pos,-4),
                self.abs_offset + Vector2(self.parent.hue_pos, 12),
                3
            )
def hex_to_rgb(hex_code):
    hex_code = hex_code[1:]
    return tuple(int(hex_code[i:i+2], 16) for i in (0, 2, 4))
class UIColorInput(UITextInput):
    def __init__(self, app, pos, size, ux=None, draggable=False, multiline = False, max_length = -1, type = 2, **kwargs):
        
        super().__init__(app, pos, size, ux, draggable, multiline, max_length, type, **kwargs)
    
    def update(self):
        if self.special_keys[PG.K_RETURN].pressed:
            try:
                # 1. HEX zu RGB konvertieren (0-255)
                r, g, b = hex_to_rgb(self.text)
                
                r_norm, g_norm, b_norm = r / 255.0, g / 255.0, b / 255.0
                
                h, s, v = rgb_to_hsv(r_norm, g_norm, b_norm)


                ly = int(h * (127)) 

                x = 127 - int(s * (127))
                
                y = int((v * (127))) 
                self.parent.color_pos = (x,y)
                self.parent.line_pos = ly
                self.parent.ux_color_rect.image = color_rect(h,128)
                
            except ValueError as E:
                print(f'ERROR: {E}')
        return super().update()

class UIColorPicker(UIElement):
    def __init__(self, app, pos, **kwargs):
        ux = [
            [UXRect(-1,color=Color('#363636'),size=Vector2(144,216)),
             UXRect(-1,color=Color('#484848' if i < 1 else '#969696'),size=Vector2(144,16)),
             UXText(color=Color('#969696'),text_get_callback=lambda : "Color")] for i in range(4)
        ]
        kwargs['anchor'] = 'tl'
        super().__init__(app, pos, Vector2(144,216), UXWrapper(ux), True, **kwargs)

        self.ux_color_rect = UXImage(Vector2(0,0),color_rect(0.2, 128),True)
        self.ux_color_line = UXImage(Vector2(0,0),color_line(Vector2(128,8)),True)
        
        self.color = Color(0,0,0)
        self.hue = 0.2
        self.hue_pos = 0
        self.color_pos = Vector2(0,0)
        
        self.color_rect_btn = UIColorPickerRect(
            self.app,
            Vector2(8,24),
            Vector2(128,128),
            UXWrapper([[self.ux_color_rect] for i in range(4)]),
            parent = self,
            anchor='tl',
            cb_lclick = self.get_color #CAlculate the mouse_pos - offset
        )
        self.color_line_btn = UIColorPickerLine(
            self.app,
            Vector2(8,160),
            Vector2(128,8),
            UXWrapper([[self.ux_color_line] for i in range(4)]),
            parent = self,
            anchor='tl',
            cb_lclick = self.get_hue #CAlculate the mouse_pos - offset
        )
        
        self.color_in_out = UIColorInput(
            self.app,
            Vector2(8,176),
            Vector2(128,32),
            max_length=7,
            type=3,
            parent = self,
            anchor = 'tl'
        )
    def get_color_hex(self) -> str:
        c = Color(self.color)
        return f'#{c.r:02x}{c.g:02x}{c.b:02x}'
        
    def get_hue(self, obj):
        self.hue_pos = obj.get_internal_mouse_pos.x
        self.hue = 255 / self.hue_pos
        self.ux_color_rect.image = color_rect(self.hue, 128)
        
        self.color = Color(self.ux_color_rect.image.get_at(self.color_pos))
        c = self.color
        self.color_in_out.text = f'#{c.r:02x}{c.g:02x}{c.b:02x}'
        
    def get_color(self, obj):
        self.color_pos = obj.get_internal_mouse_pos
        self.color = Color(self.ux_color_rect.image.get_at(self.color_pos))
        c = self.color
        self.color_in_out.text = f'#{c.r:02x}{c.g:02x}{c.b:02x}'


from pig_ui.ui.ui_element import UIElement
from pig_ui.ui.ui_dropdown import UIDropDown
from pig_ui.constants import *
from pig_ui.ui.ux_element import UXWrapper, UXText, UXRect
from pig_ui.ui.ui_manager import UIM
class UIMenuBar(UIElement):
    def __init__(self, app, pos, size, ux = None, draggable = False, **kwargs):
        kwargs['visible'] = False
        kwargs['anchor'] = 'tl'
        super().__init__(app, pos, size, ux, draggable, **kwargs)
        self.sub = []
    def set_subs(self, subs: list[dict]):
        """
        Format:
            {
                'title': "title",
                'ltext': [],
                'lcom': []
            }
        """
        
        for i, dd in enumerate(subs):
            uidd = UIDropDown(
                self.app, 
                Vector2(i*self.size.x, 0),
                self.size,
                parent=self,
                anchor= 'tl',
                title = dd['title'],
                draggable=True,
                )
            
            uidd.set_subs(dd['ltext'], dd['lcom'])
            self.sub.append(uidd)


from pig_ui.ui.ui_element import UIElement
from pig_ui.ui.ui_textinput import UITextInput
from pig_ui.ui.ux_element import UXWrapper, UXText, UXRect
from pig_ui.constants import *
class UISpinBox(UIElement): 
    def __init__(self, app, pos,increment_by: float | int,round_by: int, **kwargs):
        kwargs['anchor'] = 'tl'
        self.increment_by = increment_by
        self.round_by = round_by
        super().__init__(app, pos, Vector2(80,16), None, False, **kwargs)
        
        
        ux = [
            [UXRect(-1,Color('#242424' if i < 1 else '#484848'),size=Vector2(16,16)),
                UXText(color=Color('#969696'),text_get_callback=lambda: '<')] for i in range(4)
        ]
        
        
        self.dec_btn = UIElement(
            self.app,
            Vector2(0,0),
            Vector2(16,16),
            UXWrapper(ux),
            parent = self,
            anchor = 'tl',
            cb_lclick = self.dec_num
        )
        
        self.inp = UITextInput(
            self.app,
            Vector2(16,0),
            Vector2(48,16),
            parent = self,
            anchor = 'tl',
            type=2
        )
        
        ux = [
            [UXRect(-1,Color('#242424' if i < 1 else '#484848'),size=Vector2(16,16)),
                UXText(color=Color('#969696'),text_get_callback=lambda: '>')] for i in range(4)
        ]
        self.inc_btn = UIElement(
            self.app,
            Vector2(64,0),
            Vector2(16,16),
            UXWrapper(ux),
            anchor = 'tl',
            parent = self,
            cb_lclick = self.inc_num
        )
    def get_n(self) -> float | int:
        try:
            n = float(self.inp.text)
            return n
        except:
            return 0
    def inc_num(self, x):
        n = self.get_n()
        n += self.increment_by
        n = round(n, self.round_by)
        self.inp.text = str(n)
    
    def dec_num(self, x):
        n = self.get_n()
        n -= self.increment_by
        n = round(n, self.round_by)
        self.inp.text = str(n)


from pig_ui.constants import *

class UXParameterError(Exception):
    """"""

class UX4Param:
    def __init__(self,*options):
        options = list(options)
        l = 4 - len(options)
        if l < 0:
            raise Exception('TooManyParameters')
        if l > 0:
            options += [0 for i in range(l)]
        
        self.options = options

class UXElement:
    def draw(self): ...

class UXRect(UXElement):
    def __init__(self,
                 border_radius: int | UX4Param = -1,
                 color: Color = Color('#252525'),
                 offset: Vector2 = Vector2(0,0),
                 size: Vector2 = Vector2(1,1),
                 width: int = 0):
        self.border_radius = border_radius
        if isinstance(self.border_radius, int):
            self.border_radius = UX4Param(*[self.border_radius]*4)
        self.color = color
        self.pos = offset
        self.size = size
        self.width = width
        
    def draw(self, surf: Surface, offset: Vector2):
        PG.draw.rect(surf, 
                     self.color, 
                     (offset.x + self.pos.x,offset.y + self.pos.y, self.size.x,self.size.y),
                     self.width,
                     border_radius=-1,
                     #*self.border_radius.options
                     )

class UXCircle(UXElement): 
    def __init__(self,
                 color: Color,
                 center: Vector2,
                 radius: float,
                 width: int,
                 drawing_points: UX4Param):
        self.color = color
        self.center = center
        self.radius = radius
        self.width = width
        self.drawing_points = drawing_points
        
    def draw(self, surf):
        PG.draw.circle(
            surf,
            self.color,
            self.center,
            self.radius,
            self.width,
            *self.drawing_points.options
        )

class UXLine(UXElement):
    def __init__(self,
                 color: Color,
                 start: Vector2,
                 end: Vector2,
                 width: int):
        self.color = color
        self.start = start
        self.end = end
        self.width = width
    def draw(self, surf):
        PG.draw.line(
            surf,
            self.color,
            self.start,
            self.end,
            self.width
        )

class UXPolygon(UXElement):
    def __init__(self,
                 color: Color,
                 points: list[Vector2],
                 width: int):
        self.color = color
        self.points = points
        self.width = width
    def draw(self, surf):
        PG.draw.polygon(
            surf,
            self.color,
            self.points,
            self.width
        )

class UXImage(UXElement):
    def __init__(self,
                 pos: Vector2,
                 path: str | Surface,
                 alpha: bool):
        self.pos = pos
        if isinstance(path, str):
            self.image = PG.image.load(path)
        elif isinstance(path, Surface):
            self.image = path
            
        self.alpha = alpha
    def update(self, surf: Surface):
        self.image = surf
    def draw(self, surf, offset):
        surf.blit(self.image, self.pos+offset)

class UXText(UXElement):
    # + Add a snap point for x, so the text will not go outside the element!
    def __init__(self, pos: Vector2 = Vector2(0,0), color: Color = Color('#dddddd'), anchor = 0,text_get_callback: Callable | str = lambda: ""):
        self.anchor = anchor
        self.pos = pos
        self.text_get_callback = text_get_callback
        self.color = color
        super().__init__()
    @property
    def anchor_offset(self) -> Vector2:
        return [0,0.5,1][self.anchor]
    def draw(self, surf: Surface, offset: Vector2):
        text = self.text_get_callback() if not isinstance(self.text_get_callback, str) else self.text_get_callback
        rendered = NORM_FONT.render(text, True, self.color)
        size = Vector2(*rendered.get_size())
        surf.blit(rendered, self.pos + offset - (size * self.anchor_offset))
        
        

UIELEMENT_DEFAULT = [
    [UXRect(-1,Color('#484848'),size=Vector2(15,15))],
    [UXRect(-1,Color('#969696'),size=Vector2(15,15))],
    [UXRect(-1,Color('#ffffff'),size=Vector2(15,15))],
    [UXRect(-1,Color('#000000'),size=Vector2(15,15))]
]


UIELEMENT_TEXT = [
    [UXText(color=Color('#484848'))],
    [UXText(color=Color('#969696'))],
    [UXText(color=Color('#ffffff'))],
    [UXText(color=Color('#000000'))]
]

class UXRenderer:
    def __init__(self,
                 ui,
                 ux: list[list[UXElement]]):
        self.ui = ui
        self.ux = ux
        self.draw()
    def draw(self, surf: Surface, offset: Vector2):
        for element in self.ux:
            
            element.draw(surf, offset)
            
class UXWrapper:
    def __init__(self, ux: list[list[UXRenderer]]):
        
        self.ux = ux
        self.set_mode(0)
    
    def draw(self,surf: Surface, offset: Vector2):
        for ux in self.ux[self.selected]:
            ux.draw(surf, offset)
        
    def set_mode(self,type: int):
        """
        (0) normal
        (1) hover
        (2) click
        (3) disabled
        """
        self.selected = type
        
        
